def parse_from_readme():
    """
    Adapted from https://github.com/boxed/pytest-readme

    Copyright (c) 2020, Anders Hovmöller

    Under the BSD 3-Clause "New" or "Revised" License.
    """
    with open(
        "tests/test_readme.py",
        "w",
        encoding="utf8",
    ) as out, open("README.md", encoding="utf8") as readme:
        output, mode = [], None

        for i, line in enumerate(readme.readlines()):
            output.append("\n")

            if mode is None and line.strip() == "```python":
                mode = "first_line"
                output[i] = "def test_line_%s():\n" % i
                continue

            if line.strip() == "```":
                if mode == "doctest":
                    output[i] = '    """\n'
                mode = None
                continue

            if mode == "first_line":
                if line.strip() == "":
                    mode = None
                    output[i - 1] = "\n"
                    continue

                if line.strip().startswith(">>>"):
                    mode = "doctest"
                    output[i - 2] = (
                        output[i - 1][:-1] + "  " + output[i - 2]
                    )  # move the def line one line up
                    output[i - 1] = '    """\n'
                else:
                    mode = "test"

            if mode in ("doctest", "test"):
                output[i] = "    " + line
            else:
                output[i] = "# %s" % line

        out.writelines(output)


parse_from_readme()



from .general_sam import (
    GeneralSAM,
    GeneralSAMState,
    GreedyTokenizer,
    Trie,
    TrieNode,
)
from .trie_utils import (
    CountInfo,
    SortResult,
    build_trie_from_bytes,
    build_trie_from_chars,
    sort_bytes,
    sort_chars,
    sort_seq_via_trie,
)
from .vocab_prefix import (
    VocabPrefixAutomaton,
    VocabPrefixBytesOrChars,
)

__all__ = [
    "GeneralSAM",
    "GeneralSAMState",
    "GreedyTokenizer",
    "Trie",
    "TrieNode",
    "CountInfo",
    "SortResult",
    "build_trie_from_chars",
    "build_trie_from_bytes",
    "sort_chars",
    "sort_bytes",
    "sort_seq_via_trie",
    "VocabPrefixAutomaton",
    "VocabPrefixBytesOrChars",
]



from typing import Collection, Iterable, Optional, Sequence, Union

from general_sam import (
    CountInfo,
    GeneralSAMState,
    VocabPrefixAutomaton,
    VocabPrefixBytesOrChars,
)


def _test_token_healing_batch(
    vocab: Collection[Union[str, bytes]],
    token_sequences: Iterable[Union[Sequence[str], Sequence[bytes]]],
    bytes_or_chars: VocabPrefixBytesOrChars,
):
    automaton = VocabPrefixAutomaton(vocab, bytes_or_chars=bytes_or_chars)

    vocab_sorted = sorted(vocab)

    def validate(
        query: Union[str, bytes], state: GeneralSAMState, cnt_info: Optional[CountInfo]
    ):
        import bisect

        expected_l = bisect.bisect_left(
            vocab_sorted, query, key=lambda x: x[: len(query)]
        )
        expected_r = bisect.bisect_right(
            vocab_sorted, query, key=lambda x: x[: len(query)]
        )

        if expected_l < expected_r:
            expected_cnt_info = CountInfo(
                str_cnt=expected_r - expected_l,
                tot_cnt_lower=expected_l,
                tot_cnt_upper=expected_r,
            )
        else:
            expected_cnt_info = None

        assert cnt_info == expected_cnt_info, (query, cnt_info, expected_cnt_info)

        assert state.is_nil() ^ any(query in i for i in vocab)  # pyright: ignore

    def check(tokens: Sequence[Union[str, bytes]]):
        state = automaton.get_root_state()
        query = "" if isinstance(tokens[0], str) else b""

        # NOTE: tokens are prepended in the reverse order
        for token in reversed(tokens):
            query = token + query  # pyright: ignore
            cnt_info = automaton.prepend_feed(state, token)
            validate(query, state, cnt_info)

    for tokens in token_sequences:
        check(tokens)


def _test_batch(
    vocab: Collection[str],
    token_sequences: Iterable[Union[Sequence[str], Sequence[bytes]]],
):
    _test_token_healing_batch(
        vocab,
        tuple(filter(lambda x: isinstance(x[0], str), token_sequences)),
        VocabPrefixBytesOrChars.CHARS,
    )
    _test_token_healing_batch(
        tuple(i.encode() for i in vocab),
        tuple(
            tuple(i.encode() if isinstance(i, str) else i for i in s)
            for s in token_sequences
        ),
        VocabPrefixBytesOrChars.BYTES,
    )


def test_simple_token_healing():
    _test_batch(
        ["bb", "ca", "ab", "c", "aa", "bbaa", "a", "cc", "b"],
        [
            ["bb", "a"],
            ["b", "b", "b"],
            ["b", "b", "a"],
            ["b", "ba"],
            ["ca", "c", "ab"],
            ["c", "c", "c"],
        ],
    )


def test_simple_chinese_token_healing():
    _test_batch(
        ["歌曲", "聆听歌曲", "播放歌曲", "歌词", "查看歌词"],
        [
            ["歌曲"],
            ["聆听歌曲"],
            ["聆听", "歌曲"],
            ["聆", "听", "歌曲"],
            ["播放歌曲"],
            ["播", "放歌曲"],
            ["播放", "歌曲"],
            ["歌词"],
            ["查看歌词"],
            ["查看", "歌词"],
            ["听歌曲"],
            ["听", "歌曲"],
            ["放歌曲"],
            ["听歌"],
            ["放歌"],
            ["词"],
            ["查看"],
            ["bb", "a"],
            ["b", "b", "b"],
            ["b", "b", "a"],
            ["b", "ba"],
            ["ca", "c", "ab"],
            ["c", "c", "c"],
        ],
    )


def test_simple_utf8_token_healing():
    # '䨻'.encode('utf8') == b'\xe4\xa8\xbb'
    _test_batch(
        ["䨻"],
        [
            ["䨻"],
            [b"\xe4", b"\xa8", b"\xbb"],
            [b"\xe4", b"\xa8\xbb"],
            [b"\xe4\xa8", b"\xbb"],
            [b"\xe4\xa8\xbb"],
        ],
    )



from typing import Callable, Mapping, Optional, Sequence, Tuple, Union

ByteOrChar = Union[str, int]
TrieNodeID = int
GeneralSAMNodeID = int

NIL_NODE_ID = 0
ROOT_NODE_ID = 1

class TrieNode:
    def is_in_chars(self) -> bool: ...
    def is_in_bytes(self) -> bool: ...
    def get_node_id(self) -> TrieNodeID: ...
    def is_accepting(self) -> bool: ...
    def get_trans(self) -> Mapping[ByteOrChar, TrieNodeID]: ...
    def get_parent(self) -> TrieNodeID: ...

class Trie:
    @staticmethod
    def in_chars() -> "Trie": ...
    @staticmethod
    def in_bytes() -> "Trie": ...
    def is_in_chars(self) -> bool: ...
    def is_in_bytes(self) -> bool: ...
    def num_of_nodes(self) -> int: ...
    def insert_chars(self, s: str) -> TrieNodeID: ...
    def insert_bytes(self, s: bytes) -> TrieNodeID: ...
    def get_bfs_order(self) -> Sequence[TrieNodeID]: ...
    def get_root(self) -> TrieNode: ...
    def get_node(self, node_id: TrieNodeID) -> Optional[TrieNode]: ...
    def dfs_travel(
        self,
        in_stack_callback: Callable[[TrieNodeID, Optional[ByteOrChar]], None],
        out_stack_callback: Callable[[TrieNodeID], None],
        root_node_id: Optional[TrieNodeID] = None,
    ) -> TrieNode: ...
    def bfs_travel(
        self,
        in_queue_callback: Callable[[TrieNodeID, Optional[ByteOrChar]], None],
        out_queue_callback: Callable[[TrieNodeID], None],
        root_node_id: Optional[TrieNodeID] = None,
    ) -> TrieNode: ...

class GeneralSAMState:
    def is_in_chars(self) -> bool: ...
    def is_in_bytes(self) -> bool: ...
    def get_node_id(self) -> GeneralSAMNodeID: ...
    def is_nil(self) -> bool: ...
    def is_root(self) -> bool: ...
    def is_accepting(self) -> bool: ...
    def get_trans(self) -> Mapping[ByteOrChar, GeneralSAMNodeID]: ...
    def get_suffix_parent_id(self) -> GeneralSAMNodeID: ...
    def copy(self) -> "GeneralSAMState": ...
    def goto_suffix_parent(self) -> None: ...
    def goto_char(self, t: str) -> None: ...
    def goto_byte(self, t: int) -> None: ...
    def feed_chars(self, s: str) -> None: ...
    def feed_bytes(self, s: bytes) -> None: ...
    def dfs_along(
        self,
        trie: Trie,
        in_stack_callback: Callable[
            ["GeneralSAMState", TrieNodeID, Optional[ByteOrChar]], None
        ],
        out_stack_callback: Callable[["GeneralSAMState", TrieNodeID], None],
        trie_node_id: Optional[TrieNodeID] = None,
    ) -> TrieNode: ...
    def bfs_along(
        self,
        trie: Trie,
        in_queue_callback: Callable[
            ["GeneralSAMState", TrieNodeID, Optional[ByteOrChar]], None
        ],
        out_queue_callback: Callable[["GeneralSAMState", TrieNodeID], None],
        trie_node_id: Optional[TrieNodeID] = None,
    ) -> TrieNode: ...

class GeneralSAM:
    @staticmethod
    def from_chars(s: str) -> "GeneralSAM": ...
    @staticmethod
    def from_bytes(s: bytes) -> "GeneralSAM": ...
    @staticmethod
    def from_trie(trie: Trie) -> "GeneralSAM": ...
    def is_in_chars(self) -> bool: ...
    def is_in_bytes(self) -> bool: ...
    def num_of_nodes(self) -> int: ...
    def get_root_state(self) -> GeneralSAMState: ...
    def get_state(self, node_id: GeneralSAMNodeID) -> GeneralSAMState: ...
    def get_topo_and_suf_len_sorted_states(self) -> Sequence[GeneralSAMState]: ...

class GreedyTokenizer:
    @staticmethod
    def from_sam_and_trie(sam: GeneralSAM, trie: Trie) -> "GreedyTokenizer": ...
    def get_sam(self) -> GeneralSAM: ...
    def is_in_chars(self) -> bool: ...
    def is_in_bytes(self) -> bool: ...
    def tokenize_str(
        self, s: str, unk_token_id: TrieNodeID = NIL_NODE_ID
    ) -> Sequence[Tuple[TrieNodeID, int]]: ...
    def tokenize_bytes(
        self, s: bytes, unk_token_id: TrieNodeID = NIL_NODE_ID
    ) -> Sequence[Tuple[TrieNodeID, int]]: ...



from general_sam import GeneralSAM, GeneralSAMState, build_trie_from_chars


def test_bytes_abcbc():
    sam = GeneralSAM.from_bytes(b"abcbc")
    assert sam.is_in_bytes()

    state = sam.get_root_state()
    state.feed_bytes(b"cbc")
    assert state.is_accepting()

    state = sam.get_root_state()
    state.feed_bytes(b"bcb")
    assert not state.is_accepting()


def test_chars_abcbc():
    sam = GeneralSAM.from_chars("abcbc")
    assert sam.is_in_chars()

    state = sam.get_root_state()

    state.feed_chars("b")
    assert not state.is_accepting()
    state.feed_chars("c")
    assert state.is_accepting()
    state.feed_chars("bc")
    assert state.is_accepting()
    state.feed_chars("bc")
    assert not state.is_accepting() and state.is_nil()


def test_simple_sam_from_trie():
    trie, _ = build_trie_from_chars(["hello", "Chielo"])
    sam = GeneralSAM.from_trie(trie)
    assert trie.is_in_chars() and sam.is_in_chars()

    def fetch_state(s: str) -> GeneralSAMState:
        state = sam.get_root_state()
        state.feed_chars(s)
        return state

    assert fetch_state("lo").is_accepting()
    assert fetch_state("ello").is_accepting()
    assert fetch_state("elo").is_accepting()

    state = fetch_state("el")
    assert not state.is_accepting() and not state.is_nil()

    state = fetch_state("bye")
    assert not state.is_accepting() and state.is_nil()



from general_sam import (
    GeneralSAM,
    GreedyTokenizer,
    build_trie_from_bytes,
    build_trie_from_chars,
)


def test_english_chars_tokenize():
    vocab = ["a", "ab", "b", "bc", "c", "d", "e", "f", "cd", "abcde"]
    trie, token_to_trie_node = build_trie_from_chars(vocab)

    trie_node_to_token = [-1] * trie.num_of_nodes()
    for i, j in enumerate(token_to_trie_node):
        trie_node_to_token[j] = i

    sam = GeneralSAM.from_trie(trie)
    tokenizer = GreedyTokenizer.from_sam_and_trie(sam, trie)
    assert tokenizer.is_in_chars()

    def tokenize(s: str):
        return [(trie_node_to_token[i], j) for i, j in tokenizer.tokenize_str(s)]

    assert tokenize("abcde") == [(9, 5)]
    assert tokenize("abcdf") == [(1, 2), (8, 2), (7, 1)]
    assert tokenize("abca") == [(1, 2), (4, 1), (0, 1)]


def test_chinese_chars_tokenize():
    vocab = ["歌曲", "聆听歌曲", "播放歌曲", "歌词", "查看歌词", "听歌", "曲折"]
    trie, token_to_trie_node = build_trie_from_chars(vocab)

    trie_node_to_token = [-1] * trie.num_of_nodes()
    for i, j in enumerate(token_to_trie_node):
        trie_node_to_token[j] = i

    sam = GeneralSAM.from_trie(trie)
    tokenizer = GreedyTokenizer.from_sam_and_trie(sam, trie)
    assert tokenizer.is_in_chars()

    def tokenize(s: str):
        return [(trie_node_to_token[i], j) for i, j in tokenizer.tokenize_str(s)]

    assert tokenize("歌曲折") == [(0, 2), (-1, 1)]
    assert tokenize("听歌曲") == [(5, 2), (-1, 1)]
    assert tokenize("听歌曲折") == [(5, 2), (6, 2)]
    assert tokenize("聆听歌曲折") == [(1, 4), (-1, 1)]
    assert tokenize("查看歌词歌曲") == [(4, 4), (0, 2)]
    assert tokenize("一起播放歌曲并共享歌词") == [(-1, 2), (2, 4), (-1, 3), (3, 2)]


def test_chinese_bytes_tokenize():
    vocab = ["歌曲", "聆听歌曲", "播放歌曲", "歌词", "查看歌词", "听歌", "曲折"]
    vocab = [i.encode() for i in vocab]
    trie, token_to_trie_node = build_trie_from_bytes(vocab)

    trie_node_to_token = [-1] * trie.num_of_nodes()
    for i, j in enumerate(token_to_trie_node):
        trie_node_to_token[j] = i

    sam = GeneralSAM.from_trie(trie)
    tokenizer = GreedyTokenizer.from_sam_and_trie(sam, trie)
    assert tokenizer.is_in_bytes()

    def tokenize_str(s: str):
        return [trie_node_to_token[i] for i, _ in tokenizer.tokenize_str(s)]

    def tokenize_bytes(s: str):
        return [trie_node_to_token[i] for i, _ in tokenizer.tokenize_bytes(s.encode())]

    def tokenize(s: str):
        a = tokenize_str(s)
        b = tokenize_bytes(s)
        assert a == b
        return a

    assert tokenize("歌曲折") == [0, -1]
    assert tokenize("听歌曲") == [5, -1]
    assert tokenize("听歌曲折") == [5, 6]
    assert tokenize("聆听歌曲折") == [1, -1]
    assert tokenize("查看歌词歌曲") == [4, 0]
    assert tokenize("一起播放歌曲并共享歌词") == [-1, 2, -1, 3]
